/*
Phantasy Star Online Blue Burst Quest Disassembler

Copyright (C) 2022 Terry Chatman

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License version 3
as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <windows.h>
#include <string.h>
#include "prs.h"

//#define NO_ARGS

FILE* stringFile;

typedef struct {
	long val;
	wchar_t name[256];
	wchar_t mask[256];
} psoOpcode;

int importStrings = 0;
int numOpcodes = 0;
wchar_t replaceBuf[4096] = { 0 };
wchar_t stringBuf[4096] = { 0 };
wchar_t asmFileName[4096] = { 0 };

psoOpcode* opcode_list[1024];

typedef struct {
	const wchar_t* arg;
	unsigned size;
} psoOpcodeArg;

psoOpcodeArg opcodeSizes[] =
{ { L"T_NONE", 0x00 },
{ L"T_IMED", 0x00 },
{ L"T_ARGS", 0x10000 },
{ L"T_VASTART", 0x00 },
{ L"T_VAEND", 0x00 },
{ L"T_STR", 0x01 },
{ L"T_SWITCH", 0x01 },
{ L"T_SWITCH2B", 0x01 },
{ L"T_PUSH", 0x00 },
{ L"T_REG", 0x01 },
{ L"T_BREG", 0x01 },
{ L"T_BYTE", 0x01 },
{ L"T_WORD", 0x01 },
{ L"T_FUNC", 0x01 },
{ L"T_FUNC2", 0x01 },
{ L"T_PFLAG", 0x01 },
{ L"T_DATA", 0x01 },
{ L"T_STRDATA", 0x01 },
{ L"T_DWORD", 0x01 },
{ L"T_FLOAT", 0x01 },
{ L"T_DC", 0x00 },
{ L"T_V2", 0x00 },
{ L"T_V3", 0x00 },
{ L"T_V4", 0x00 },
};

long getMaskSize(const wchar_t* mask)
{
	unsigned maskSize = 0;
	wchar_t* t = NULL;
	wchar_t* nt1 = NULL;
	unsigned ch;
	psoOpcodeArg* a;
	wchar_t maskproc[4096];

	if (mask == NULL)
		return -1;

	swprintf(maskproc, 4096, mask);

	t = wcstok(maskproc, L", ", &nt1);

	while (t != NULL)
	{
		for (ch = 0; ch < sizeof(opcodeSizes) / sizeof(psoOpcodeArg); ch++)
		{
			a = &opcodeSizes[ch];
			if (wcscmp(t, a->arg) == 0)
			{
				switch (a->size)
				{
				case 0x10000: // T_ARGS
					return 0x10000;
					break;
				default:
					break;
				}
			}
		}
		t = wcstok(NULL, L", ", &nt1);
	}
	return 0;
}

void parseOpcodeList()
{
	FILE* of;
	char opCodeData[1024];
	unsigned ch;
	char* t;
	unsigned ch2, ch3;
	psoOpcode* o;
	psoOpcodeArg* a;
	int Dreamcast;
	int functionOK;
	wchar_t wcsbuf[256];
	wchar_t* w;

	of = _wfopen(asmFileName, L"r");
	if (!of)
	{
		printf("FATAL: asm.txt is missing!");
		exit(1);
	}
	else
	{
		while (fgets(opCodeData, 1024, of) != NULL)
		{
			if (strlen(opCodeData) > 1)
			{
				for (ch = 1; ch < strlen(opCodeData); ch++)
				{
					if (opCodeData[ch] == 0x7D)
						opCodeData[ch] = 0x00;
				}
				t = strtok(&opCodeData[1], ", ");
				Dreamcast = 0;
				if (t != NULL)
				{
					ch2 = 0;
					o = calloc(1, sizeof(psoOpcode));

					if (!o)
					{
						wprintf(L"Out of memory!");
						exit(1);
					}

					while (t != NULL)
					{
						wcsbuf[0] = 0;
						w = &wcsbuf[0];

						for (ch3 = 0; ch3 < strlen(t); ch3++)
						{
							if (t[ch3] != 0x22)
								* w++ = t[ch3];
						}

						*w++ = 0;
						w = &wcsbuf[0];

						switch (ch2)
						{
						case 0: // opcode value
							o->val = (long)wcstoll(w, NULL, 0);
							break;
						case 1:
							swprintf_s(o->name, 256, w);
							break;
						default:
							if (wcslen(o->mask))
								wcscat(o->mask, L", ");
							wcscat(o->mask, w);
							if ((wcscmp(w, L"T_DC") == 0) || (wcscmp(w, L"T_DREG") == 0))
								Dreamcast = 1;
							else
							{
								functionOK = 0;
								for (ch3 = 0; ch3 < sizeof(opcodeSizes) / sizeof(psoOpcodeArg); ch3++)
								{
									a = &opcodeSizes[ch3];
									if (wcscmp(a->arg, w) == 0)
									{
										functionOK = 1;
										break;
									}
								}
								if (!functionOK)
								{
									wprintf(L"FATAL: Processing asm.txt, I don't understand \"%s\"!", w);
									exit(1);
								}
							}
							break;
						}
						t = strtok(NULL, ", ");
						ch2++;
					}
					if (!Dreamcast)
					{
						if (numOpcodes > 1024)
						{
							wprintf(L"FATAL: Processing asm.txt, too many opcodes in file!");
							exit(1);
						}						
						opcode_list[numOpcodes++] = o;
					}
					else
						free(o);
				}
			}
		}
		wprintf(L"%u opcodes loaded from asm.txt\n", (unsigned)numOpcodes);
		fclose(of);
	}
}

int CreateQSTSize(unsigned char* binPrs, unsigned binPrsSize, unsigned char* datPrs, unsigned datPrsSize)
{
	unsigned ch = 0;
	unsigned ch2 = 0;
	unsigned ch3 = 0xB0;
	unsigned ch4 = 0;
	unsigned ch5 = 0;

	while ((ch < datPrsSize) || (ch2 < binPrsSize))
	{
		if (ch < datPrsSize)
		{
			ch4 = datPrsSize - ch;
			if (ch4 > 0x400)
				ch4 = 0x400;
			ch += ch4;
			ch3 += 0x420;
		}
		if (ch2 < binPrsSize)
		{
			ch4 = binPrsSize - ch2;
			if (ch4 > 0x400)
				ch4 = 0x400;
			ch2 += ch4;
			ch3 += 0x420;
		}
		ch5++;
	}
	return ch3;
}

int CreateQST(unsigned char* qp, unsigned char* binPrs, unsigned binPrsSize, unsigned char* datPrs, unsigned datPrsSize, unsigned short qnum)
{
	unsigned ch = 0;
	unsigned ch2 = 0;
	unsigned ch3 = 0xB0;
	unsigned ch4 = 0;
	unsigned ch5 = 0;

	*(unsigned*)& qp[0x00] = 0x00440058;
	*(unsigned short*)& qp[0x04] = qnum;
	sprintf(&qp[0x2C], "quest%u.dat", qnum);
	*(unsigned*)& qp[0x3C] = datPrsSize;
	sprintf(&qp[0x40], "quest%u_j.dat", qnum);
	*(unsigned*)& qp[0x58] = 0x00440058;
	*(unsigned short*)& qp[0x5C] = qnum;
	sprintf(&qp[0x84], "quest%u.bin", qnum);
	*(unsigned*)& qp[0x94] = binPrsSize;
	sprintf(&qp[0x98], "quest%u_j.bin", qnum);

	while ((ch < datPrsSize) || (ch2 < binPrsSize))
	{
		if (ch < datPrsSize)
		{
			ch4 = datPrsSize - ch;
			if (ch4 > 0x400)
				ch4 = 0x400;
			*(unsigned*)& qp[ch3] = 0x0013041C;
			*(unsigned*)& qp[ch3 + 0x04] = ch5;
			sprintf(&qp[ch3 + 0x08], "quest%u.dat", qnum);
			memcpy(&qp[ch3 + 0x18], &datPrs[ch], ch4);
			*(unsigned*)& qp[ch3 + 0x418] = ch4;
			ch += ch4;
			ch3 += 0x420;
		}
		if (ch2 < binPrsSize)
		{
			ch4 = binPrsSize - ch2;
			if (ch4 > 0x400)
				ch4 = 0x400;
			*(unsigned*)& qp[ch3] = 0x0013041C;
			*(unsigned*)& qp[ch3 + 0x04] = ch5;
			sprintf(&qp[ch3 + 0x08], "quest%u.bin", qnum);
			memcpy(&qp[ch3 + 0x18], &binPrs[ch2], ch4);
			*(unsigned*)& qp[ch3 + 0x418] = ch4;
			ch2 += ch4;
			ch3 += 0x420;
		}
		ch5++;
	}
	return ch3;
}

wchar_t* replaceStr(const wchar_t* s)
{
	int c = 0, d = 0;
	int l = wcslen(s);
	wchar_t checkbuf[10];
	int match;

	replaceBuf[0] = 0;

	if (importStrings)
	{
		if (fgetws(stringBuf, 4096, stringFile) == NULL)
		{
			wprintf(L"WARNING: Ran out of strings to import!\n");
			return replaceBuf;
		}
		while (c < l)
		{
			if (stringBuf[c] == 0x25)
			{
				replaceBuf[d++] = 0x25;
				replaceBuf[d++] = 0x25;
				c++;
			}
			else
			{
				match = 0;
				if (l - c >= 3)
				{
					memcpy(checkbuf, &stringBuf[c], 3 * 2);
					checkbuf[3] = 0;
					if (!wcscmp(checkbuf, L"<t>"))
					{
						replaceBuf[d++] = 0x09;
						c += 3;
						match = 1;
					}
				}
				if ((l - c >= 4) && (!match))
				{
					memcpy(checkbuf, &stringBuf[c], 4 * 2);
					checkbuf[4] = 0;
					if (!wcscmp(checkbuf, L"<cr>"))
					{
						replaceBuf[d++] = 0x0A;
						c += 4;
						match = 1;
					}
				}
				if (!match)
				{
					if ((stringBuf[c] != 0x0A) && (stringBuf[c] != 0x0D))
						replaceBuf[d++] = stringBuf[c++];
					else
						c++;
				}
			}
		}
	}
	else
	{
		while (c < l)
		{
			if (s[c] == 0x25)
			{
				replaceBuf[d++] = 0x25;
				replaceBuf[d++] = 0x25;
				c++;
			}
			else
			{
				match = 0;
				if (l - c >= 3)
				{
					memcpy(checkbuf, &s[c], 3 * 2);
					checkbuf[3] = 0;
					if (!wcscmp(checkbuf, L"<t>"))
					{
						replaceBuf[d++] = 0x09;
						c += 3;
						match = 1;
					}
				}
				if ((l - c >= 4) && (!match))
				{
					memcpy(checkbuf, &s[c], 4 * 2);
					checkbuf[4] = 0;
					if (!wcscmp(checkbuf, L"<cr>"))
					{
						replaceBuf[d++] = 0x0A;
						c += 4;
						match = 1;
					}
				}
				if (!match)
				{
					if ((s[c] != 0x0A) && (s[c] != 0x0D))
						replaceBuf[d++] = s[c++];
					else
						c++;
				}
			}
		}
	}
	replaceBuf[d++] = 0;
	return replaceBuf;
}

int ProcessMask(unsigned char* sb, psoOpcode* opCode, wchar_t* argptrs[], int argcount, wchar_t* command, unsigned lineCount, int usingArgs)
{
	int cv;
	float cf;
	int sbOfs = 0;
	wchar_t maskproc[4096];
	wchar_t* t = NULL;
	wchar_t* nt1 = NULL;
	wchar_t* carg;
	int c;
	unsigned ch, ch2;
	wchar_t* substart;
	wchar_t subbuf[4096];
	wchar_t* subptrs[256];
	wchar_t* sarg;
	int subcount = 0;
	int isRegister;

	swprintf_s(maskproc, 4096, opCode->mask);

	t = wcstok(maskproc, L", ", &nt1);
	ch = 0;

	while (t != NULL)
	{
		for (c = 0; c < sizeof(opcodeSizes) / sizeof(psoOpcodeArg); c++)
		{
			if ((wcscmp(t, opcodeSizes[c].arg) == 0) && (opcodeSizes[c].size & 0xFFFF))
			{
				if (argptrs[ch])
				{
					carg = argptrs[ch];

					if ((carg[0] == 'R') && (wcscmp(t, L"T_STR")))
						isRegister = 1;
					else
						isRegister = 0;

					if (!isRegister)
					{
						if ((wcscmp(t, L"T_FUNC") == 0) || (wcscmp(t, L"T_FUNC2") == 0) || (wcscmp(t, L"T_DATA") == 0) || (wcscmp(t, L"T_STRDATA") == 0))
							cv = _wtoi(carg);
						else
							if (wcscmp(t, L"T_FLOAT") == 0)
							{
								cf = (float)_wtof(carg);
								cv = *(unsigned*)& cf;
							}
							else
								cv = (unsigned long)wcstoll(carg, NULL, 16);
					}
					else
					{
						cv = _wtoi(carg + 1);
						if ((cv < 0) || (cv > 255))
						{
							wprintf(L"Invalid register used for opcode %s on line %u!  Register value is %i when needs to be 0..255!\n", command, lineCount, cv);
							exit(1);
						}
					}

					if ((usingArgs) && (isRegister))
					{
						if (wcscmp(t, L"T_BREG") == 0)
							sb[sbOfs++] = 0x4A;
						else
							sb[sbOfs++] = 0x48;
						sb[sbOfs++] = (unsigned char)cv;
					}
					else
						if (wcscmp(t, L"T_STR") == 0)
						{
							if (usingArgs)
								sb[sbOfs++] = 0x4E;
							memcpy(&sb[sbOfs], carg, (wcslen(carg) * 2) + 2);
							sbOfs += (wcslen(carg) * 2) + 2;
						}
						else
							if (wcscmp(t, L"T_REG") == 0)
							{
								if (usingArgs)
									sb[sbOfs++] = 0x4A;
								sb[sbOfs++] = (unsigned char)cv;
							}
							else
								if (wcscmp(t, L"T_BREG") == 0)
								{
									if (usingArgs)
										sb[sbOfs++] = 0x4A;
									sb[sbOfs++] = (unsigned char)cv;
								}
								else
									if ((wcscmp(t, L"T_FUNC") == 0) || (wcscmp(t, L"T_FUNC2") == 0) || (wcscmp(t, L"T_DATA") == 0) || (wcscmp(t, L"T_STRDATA") == 0))
									{
										if (usingArgs)
											sb[sbOfs++] = 0x4B;
										*(unsigned short*)& sb[sbOfs] = (unsigned short)cv;
										sbOfs += 2;
									}
									else
										if (wcscmp(t, L"T_BYTE") == 0)
										{
											if (usingArgs)
												sb[sbOfs++] = 0x4A;
											sb[sbOfs++] = (unsigned char)cv;
										}
										else
											if ((wcscmp(t, L"T_WORD") == 0) || (wcscmp(t, L"T_PFLAG") == 0))
											{
												if (usingArgs)
													sb[sbOfs++] = 0x4B;
												*(unsigned short*)& sb[sbOfs] = (unsigned short)cv;
												sbOfs += 2;
											}
											else
												if ((wcscmp(t, L"T_DWORD") == 0) || (wcscmp(t, L"T_FLOAT") == 0))
												{
													if (usingArgs)
														sb[sbOfs++] = 0x49;
													*(unsigned*)& sb[sbOfs] = (unsigned)cv;
													sbOfs += 4;
												}
												else
													if ((wcscmp(t, L"T_SWITCH") == 0) || (wcscmp(t, L"T_SWITCH2B") == 0))
													{
														subcount = 0;
														substart = sarg = &subbuf[0];
														*sarg = 0;
														for (ch2 = 0; ch2 < wcslen(argptrs[ch]); ch2++)
														{
															switch (argptrs[ch][ch2])
															{
															case 0x3A: // colon
																*sarg++ = 0;
																*sarg++ = 0;
																subptrs[subcount++] = substart;
																substart = sarg;
																*sarg = 0;
																break;
															default:
																*sarg++ = argptrs[ch][ch2];
																break;
															}
														}
														*sarg++ = 0;
														subptrs[subcount++] = substart;
														cv = _wtoi(subptrs[0]);
														if ((cv + 1) < subcount)
														{
															wprintf(L"Invalid amount of parameters for arguments on line %u!\n", lineCount);
															exit(1);
														}
														sb[sbOfs++] = (unsigned char)cv;
														if (wcscmp(t, L"T_SWITCH2B") == 0)
														{
															for (ch2 = 1; ch2 < (unsigned)subcount; ch2++)
															{
																cv = _wtoi(subptrs[ch2]);
																sb[sbOfs++] = (unsigned char)cv;
															}
														}
														else
														{
															for (ch2 = 1; ch2 < (unsigned)subcount; ch2++)
															{
																cv = _wtoi(subptrs[ch2]);
																*(short*)& sb[sbOfs] = cv;
																sbOfs += 2;
															}
														}
													}
					ch++;
				}
			}
		}
		t = wcstok(NULL, L", ", &nt1);
	}
	return sbOfs;
}

int wmain(int argc, wchar_t* argv[])
{
	FILE* qf;
	FILE* sf;
	wchar_t filebuf[4096];
	wchar_t command[256];
	unsigned char* sb;
	unsigned char* db;
	unsigned char* qd;
	int ds;
	int* functionTable;
	unsigned lineCount = 0;
	unsigned cl;
	int fc;
	unsigned ch, ch2;
	unsigned sbOfs;
	int highestFunction = -1;
	unsigned functionNumber;
	psoOpcode* opCode;
	int argstrings[256];
	wchar_t* argptrs[256];
	wchar_t argbuf[4096];
	wchar_t scriptFileName[4096];
	wchar_t* carg;
	wchar_t* argstart;
	int argcount = 0;
	int cv;
	wchar_t subbuf[4096];
	int gotLongDescription = 0;
	unsigned char* binPrs;
	unsigned char* datPrs;
	int binPrsSize, datPrsSize;
	int hexStrFunction = 0;

	wprintf(L"\nPhantasy Star Online Blue Burst Quest Assembler version 2.4 by Sodaboy\n");
	wprintf(L"-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n");

	GetModuleFileName(NULL, asmFileName, 4096);

	lineCount = 0;
	for (cl = 0; cl < wcslen(asmFileName); cl++)
	{
		if (asmFileName[cl] == 92)
			lineCount = cl;
	}

	if (lineCount < 4090)
		swprintf_s(&asmFileName[lineCount + 1], 4096 - lineCount, L"asm.txt");

	cl = lineCount = 0;

	parseOpcodeList();

	if (argc > 1)
		swprintf_s(scriptFileName, 4096, argv[1]);
	else
		swprintf_s(scriptFileName, 4096, L"quest");

	if (argc > 2)
	{
		if ((wcscmp(argv[2], L"-s") == 0) ||
			(wcscmp(argv[2], L"-S") == 0))
			importStrings = 1;
	}
	else
		wprintf(L"NOTE: You can use the command line parameter \"-s\" to import an\nalternate set of strings during reassembly of a quest.\n");

	if (wcslen(scriptFileName) > 5)
	{
		filebuf[5] = 0;
		for (ch = 0; ch < wcslen(scriptFileName) - 4; ch++)
		{
			memcpy(filebuf, &scriptFileName[ch], 5 * 2);
			for (ch2 = 0; ch2 < 5; ch2++)
				filebuf[ch2] = tolower(filebuf[ch2]);
			if (!wcscmp(filebuf, L".pasm"))
			{
				scriptFileName[ch] = 0;
				break;
			}
		}
	}

	sb = calloc(1, 1024768);
	functionTable = malloc(262144);

	if ((!sb) || (!functionTable))
	{
		wprintf(L"Out of memory!\n");
		return 1;
	}

	for (ch = 0; ch < 65536; ch++)
		functionTable[ch] = -1;

	swprintf_s(filebuf, 4096, L"%s.pasm.header", scriptFileName);

	_wfopen_s(&sf, filebuf, L"r, ccs=UTF-16LE");
	if (!sf)
	{
		wprintf(L"Couldn't open %s\n", filebuf);
		return 1;
	}

	if (importStrings)
	{	
		swprintf(filebuf, 32, L"%s.pasm.strings", scriptFileName);

		wprintf(L"NOTE: Will import strings from %s ...\n", filebuf);

		_wfopen_s(&stringFile, filebuf, L"r, ccs=UTF-16LE");

		if (!stringFile)
		{
			wprintf(L"Failed to open %s for writing!\n", filebuf);
			return 1;
		}
	}

	lineCount = 0;

	while (fgetws(filebuf, 4096, sf) != NULL)
	{
		switch (lineCount)
		{
		case 0: // Quest name
			swprintf_s((wchar_t*)& sb[0x18], 32, replaceStr(filebuf));
			break;
		case 1: // Quest number
			*(unsigned short*)& sb[0x10] = (unsigned short)_wtoi(filebuf);
			break;
		case 2: // Quest short description
			swprintf_s((wchar_t*)& sb[0x58], 128, replaceStr(filebuf));
			break;
		case 3: // Quest long description
			swprintf_s((wchar_t*)& sb[0x158], 256, replaceStr(filebuf));
			break;
		case 4: // Item data
			break;
		}
		lineCount++;
	}

	fclose(sf);

	swprintf_s(filebuf, 4096, L"%s.pasm", scriptFileName);

	_wfopen_s(&sf, filebuf, L"r, ccs=UTF-16LE");
	if (!sf)
	{
		wprintf(L"Couldn't open %s\n", filebuf);
		return 1;
	}

	swprintf_s(filebuf, 4096, L"%s.bin", scriptFileName);

	_wfopen_s(&qf, filebuf, L"wb");
	if (!qf)
	{
		wprintf(L"Couldn't open %s for writing!\n", filebuf);
		return 1;
	}

	sbOfs = 0x122C;

	lineCount = 0;
	*(unsigned*)& sb[0x00] = 0x122C;
	*(unsigned*)& sb[0x0C] = 0xFFFFFFFF;

	while (fgetws(filebuf, 4096, sf) != NULL)
	{
		lineCount++;
		fc = -1;
		for (ch = 0; ch < wcslen(filebuf); ch++)
		{
			switch (filebuf[ch])
			{
			case 0x20: // Space
			case 0x09: // Tab
			case 0x0A: // New Line
			case 0x0D: // Carriage Return
				break;
			default:
				fc = filebuf[ch];
				break;
			}
			if (fc != -1)
				break;
		}

		cl = 0;

		switch (fc)
		{
		case -1: // Empty line
		case 0x23: // Hash Tag Comment
		case 0x2F: // Forward Slash Comment
			break;
		default: // Command
			if (ch < wcslen(filebuf))
			{
				cl = 0;

				for (; ch < wcslen(filebuf); ch++)
				{
					switch (filebuf[ch])
					{
					case 0x3A:
						if (cl)
							memcpy(argbuf, command, cl * 2);
						argbuf[cl] = 0;
						if ((wcscmp(argbuf, L"STR")) && (wcscmp(argbuf, L"HEX")))
						{
							functionNumber = _wtoi(argbuf);
							//wprintf(L"Function %u found.\n", functionNumber);
							if (functionNumber >= 65536)
							{
								wprintf(L"Function number %u is too high.  (>= 65536)\n", functionNumber);
								return 1;
							}
							if ((functionNumber > (unsigned)highestFunction) || (highestFunction == -1))
								highestFunction = functionNumber;
							functionTable[functionNumber] = sbOfs - 0x122C;
							cl = 0;
							command[0] = 0;
							while ((filebuf[ch + 1] == 0x20) || (filebuf[ch + 1] == 0x09)) // Skip spaces and tabs
								ch++;
						}
						else
						{
							// HEX or STR function.

							ch += 2;
							fc = -1;
						}
						break;
					case 0x20:
					case 0x09:
					case 0x0A:
					case 0x0D:
						fc = -1;
						break;
					default:
						command[cl++] = filebuf[ch];
						break;
					}
					if (fc == -1)
						break;
				}

				command[cl] = 0;

				if (cl)
				{
					if ((!wcscmp(command, L"STR")) || (!wcscmp(command, L"HEX")))
					{
						if (!wcscmp(command, L"STR"))
						{
							swprintf_s((wchar_t*)& sb[sbOfs], 4096, replaceStr(&filebuf[ch]));
							sbOfs += (wcslen((wchar_t*)&sb[sbOfs]) * 2) + 2;
						}
						else
							if (!wcscmp(command, L"HEX"))
							{
								subbuf[2] = 0;
								for (; ch < wcslen(filebuf); ch += 3)
								{
									if (ch + 1 < wcslen(filebuf))
									{
										subbuf[0] = filebuf[ch];
										subbuf[1] = filebuf[ch + 1];
										cv = wcstol(&subbuf[0], NULL, 16);
										sb[sbOfs++] = cv;
									}
								}
							}
					}
					else
					{
						opCode = NULL;
						fc = -1;

						for (ch2 = 0; ch2 < (unsigned)numOpcodes; ch2++)
						{
							if (!wcscmp(opcode_list[ch2]->name, command))
							{
								opCode = opcode_list[ch2];
								fc = opCode->val;
								break;
							}
						}

						if (!opCode)
						{
							wprintf(L"Couldn't match opcode for command %s!\n", command);
							return 1;
						}

						argcount = 0;

						for (ch2 = 0; ch2 < 256; ch2++)
						{
							argptrs[ch2] = NULL;
							argstrings[ch2] = 0;
						}

						// Now split out the arguments

						ch++;
						if (ch < wcslen(filebuf))
						{
							argstart = carg = &argbuf[0];
							*carg = 0;
							for (; ch < wcslen(filebuf); ch++)
							{
								switch (filebuf[ch])
								{
								case 0x09: // Tab
									break;
								case 0x0A: // Line Break
								case 0x0D: // Carriage Return
									*carg++ = 0;
									*carg++ = 0;
									argptrs[argcount++] = argstart;
									argstart = carg;
									*carg = 0;
									ch = wcslen(filebuf);
									break;
								case 0x20: // space
									break;
								case 0x27: // apostrophe (start of string)
									ch++;
									if (ch2 = wcslen(&filebuf[ch]))
									{
										memcpy(carg, &filebuf[ch], ch2 * 2 + 2);
										ch2--;
										while (1)
										{
											if (carg[ch2] == 0x27)
											{
												carg[ch2] = 0x00;
												break;
											}
											if (ch2 == 0)
												break;
											ch2--;
										}
										replaceStr(carg);
										swprintf_s(carg, 4096 - (carg - argbuf), replaceBuf);
										carg += wcslen(carg) + 2;
									}
									ch = wcslen(filebuf); // stop processing this line as strings always occur at the end of the line
								case 0x2C: // comma
									*carg++ = 0;
									*carg++ = 0;
									argptrs[argcount++] = argstart;
									argstart = carg;
									*carg = 0;
									break;
								default:
									*carg++ = filebuf[ch];
									break;
								}
							}
						}

#ifndef NO_ARGS
						// If this function uses arguments, write out the opcodes to pass the arguments before writing out the opcode.
						if ((argcount) && (getMaskSize(opCode->mask) == 0x10000))
							sbOfs += ProcessMask(&sb[sbOfs], opCode, argptrs, argcount, command, lineCount, 1);
#endif

						if (fc >= 0xF900)
						{
							sb[sbOfs++] = 0xF9;
							sb[sbOfs++] = (unsigned char)(opCode->val & 0xFF);
						}
						else
							if (fc >= 0xF800)
							{
								sb[sbOfs++] = 0xF8;
								sb[sbOfs++] = (unsigned char)(opCode->val & 0xFF);
							}
							else
								sb[sbOfs++] = (unsigned char)(opCode->val & 0xFF);

						if ((argcount) && (getMaskSize(opCode->mask) == 0))
							sbOfs += ProcessMask(&sb[sbOfs], opCode, argptrs, argcount, command, lineCount, 0);
					}
				}
			}
			break;
		}
	}

	*(unsigned*)& sb[0x04] = sbOfs;

	if (highestFunction != -1)
	{
		memcpy(&sb[sbOfs], functionTable, (highestFunction + 1) * 4);
		sbOfs += (highestFunction + 1) * 4;
	}

	*(unsigned*)& sb[0x08] = sbOfs;

	fclose(sf);
	fwrite(sb, sbOfs, 1, qf);
	fclose(qf);

	wprintf(L"Assembly complete!\n");
	wprintf(L"%u lines read!\n", lineCount);

	free(functionTable);

	swprintf_s(filebuf, 4096, L"%s.dat", scriptFileName);

	_wfopen_s(&qf, filebuf, L"rb");
	if (qf)
	{
		wprintf(L"Reading object data from %s...\n", filebuf);
		fseek(qf, 0, SEEK_END);
		ds = ftell(qf);
		if (ds > 0)
		{
			db = malloc(ds);
			if (db != NULL)
			{
				fseek(qf, 0, SEEK_SET);
				fread(db, ds, 1, qf);
				fclose(qf);
			}
			else
			{
				wprintf(L"Out of memory!");
				exit(1);
			}
		}
		else
		{
			wprintf(L"Failed to seek into %s!", filebuf);
			exit(1);
		}

		wprintf(L"Compressing binary data...\n");

		binPrsSize = prs_compress(sb, &binPrs, sbOfs);

		if (binPrsSize < 0)
		{
			wprintf(L"Failed to compress binary data.");
			exit(1);
		}

		wprintf(L"Compressing object data...\n");

		datPrsSize = prs_compress(db, &datPrs, ds);
		if (datPrsSize < 0)
		{
			wprintf(L"Failed to compress object data.");
			exit(1);
		}

		ds = CreateQSTSize(binPrs, binPrsSize, datPrs, datPrsSize);

		qd = calloc(1, ds);

		if (!qd)
		{
			wprintf(L"Out of memory!");
			exit(1);
		}

		wprintf(L"Interleaving binary data with object data...\n");
		CreateQST(qd, binPrs, binPrsSize, datPrs, datPrsSize, *(unsigned short*)& sb[0x10]);

		swprintf_s(filebuf, 4096, L"%s.qst", scriptFileName);
		wprintf(L"Writing %s...\n", filebuf);

		_wfopen_s(&qf, filebuf, L"wb");
		if (qf)
		{
			fwrite(qd, ds, 1, qf);
			fclose(qf);
		}
		else
		{
			wprintf(L"Failed to open %s for writing!", filebuf);
			exit(1);
		}
		wprintf(L"DONE!\n");

		free(qd);
		free(datPrs);
		free(binPrs);
		free(db);
	}

	free(sb);

	if (importStrings)
		fclose(stringFile);

	return 0;
}